from datetime import timedelta
from itertools import count

from dwclib.common.db import dwcuri

import dask.dataframe as dd
from dask import delayed

one_hour = timedelta(hours=1)


def build_divisions(dtbegin, dtend, interval):
    ranges = []
    for i in count():
        beg = dtbegin + i * interval
        end = beg + interval
        ranges.append((beg, end))
        if end >= dtend:
            break
    divisions = [beg for beg, _ in ranges]
    divisions.append(dtend)
    return (ranges, divisions)


def read_data(
    runner,
    meta,
    dtbegin,
    dtend,
    uri,
    interval=one_hour,
    *args,
    **kwargs,
):
    if not uri:
        uri = dwcuri
    ranges, divisions = build_divisions(dtbegin, dtend, interval)
    parts = []
    for begin, end in ranges:
        parts.append(delayed(runner)(uri, begin, end, *args, **kwargs))
    return dd.from_delayed(parts, meta, divisions=divisions)



from collections import defaultdict
from datetime import datetime
from multiprocessing.pool import ThreadPool
from typing import List, Optional, Union

import pandas as pd
from dwclib.common.db import dwcuri
from dwclib.common.meta import waves_meta_tz
from dwclib.common.wave_unfold import unfold_row
from dwclib.common.waves import run_waves_query


def read_waves(
    patientid: str,
    dtbegin: Union[str, datetime],
    dtend: Union[str, datetime],
    labels: Optional[List[str]] = None,
    uri: str = None,
) -> pd.DataFrame:
    if not uri:
        uri = dwcuri
    if labels is None:
        labels = []
    df = run_waves_query(uri, dtbegin, dtend, patientid, labels)
    return unfold_pandas_dataframe(df)


def unfold_pandas_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    databuffer = defaultdict(list)
    for _, row in df.iterrows():
        srow = unfold_row(row, naive_datetime=False)
        databuffer[row['Label']].append(srow)
    if not databuffer:
        return waves_meta_tz
    concatter = dictconcatter(databuffer)
    with ThreadPool() as pool:
        pool.map(concatter, databuffer.keys())
    return pd.DataFrame(databuffer)


def dictconcatter(d) -> pd.DataFrame:
    def dictconcat_runner(k):
        d[k] = pd.concat(d[k], axis=0, copy=False).groupby(level=0).max()

    return dictconcat_runner



"""Python wrapper to DataWarehouse Connect."""

from dwclib.alerts import read_alerts
from dwclib.numerics import read_numerics
from dwclib.patients import read_patient, read_patients
from dwclib.waves import read_waves

__all__ = [
    'common',
    'numerics',
    'waves',
    'alerts',
    'patients',
    'dask',
    'assets',
    'read_waves',
    'read_numerics',
    'read_patients',
    'read_patient',
    'read_alerts',
]



from datetime import timedelta
from typing import List, Optional

from dwclib.common.meta import waves_meta
from dwclib.common.waves import run_waves_query
from dwclib.dask.generic import read_data

from .waves_convert import convert_dataframe

one_hour = timedelta(hours=1)


def read_waves(
    patientid,
    dtbegin,
    dtend,
    labels: Optional[List[str]] = None,
    interval: timedelta = one_hour,
    npartitions: Optional[int] = None,
    uri=None,
):
    if labels is None:
        labels = []
    ddf = read_data(
        runner=run_waves_query,
        meta=waves_meta,
        dtbegin=dtbegin,
        dtend=dtend,
        uri=uri,
        interval=interval,
        patientid=patientid,
        labels=labels,
        naive_datetime=True,
    )
    return convert_dataframe(ddf, labels, npartitions)



from functools import partial
from typing import List, Optional

import numpy as np
import pandas as pd
from dwclib.common.wave_unfold import unfold_row

import dask.dataframe as dd
from dask.dataframe.utils import make_meta


def unfold_pandas_dataframe(
    df: pd.DataFrame, columns: List[str], meta: pd.DataFrame
) -> pd.DataFrame:
    idx = pd.MultiIndex.from_arrays(
        [df.index, df['Label']], names=('TimeStamp', 'Label')
    )
    df = df.set_index(idx)
    unfolded = df.apply(unfold_row, axis=1)
    if not len(unfolded):
        return meta
    transposed = unfolded.droplevel(0).groupby(level=0).max().T
    missing_columns = set(columns) - set(transposed.columns)
    result = transposed.assign(**{c: np.nan for c in missing_columns})
    return result.reindex(sorted(result.columns), axis=1)


def convert_dataframe(
    ddf: dd.DataFrame, labels: List[str], npartitions: Optional[int] = None
) -> dd.DataFrame:
    if not labels:
        labels = ddf['Label'].unique().compute()
    labels = sorted(set(labels))
    dfmeta = pd.DataFrame(columns=labels, dtype='float32')
    idx = pd.DatetimeIndex([], name='TimeStamp')
    meta = make_meta(dfmeta, index=idx)
    if not npartitions:
        npartitions = 20 * ddf.npartitions
    ddf = ddf.repartition(npartitions)
    unfold_func = partial(unfold_pandas_dataframe, meta=meta)
    return ddf.map_partitions(unfold_func, columns=labels, meta=meta)



from configparser import ConfigParser
from pathlib import Path
from warnings import warn

from platformdirs import user_config_dir

try:
    import pyodbc

    pyodbc.pooling = False
    odbcdriver = pyodbc.drivers()[0]
except (ImportError, IndexError):
    odbcdriver = None

from sqlalchemy.engine import URL

config_dir = Path(user_config_dir(appname="dwclib", appauthor="larib-data"))
config_dir.mkdir(parents=True, exist_ok=True)
configfile = config_dir / "config.ini"


if configfile.exists():
    config = ConfigParser()
    config.read(configfile)
    dwcdriver = config.get('dwclib', 'dwc_driver', fallback=odbcdriver)
    if dwcdriver is None:
        warn('No ODBC driver found for DWC', RuntimeWarning)

    dwcuri = URL.create(
        'mssql+pyodbc',
        username=config.get('dwclib', 'dwc_user'),
        password=config.get('dwclib', 'dwc_pass'),
        host=config.get('dwclib', 'dwc_host'),
        database=config.get('dwclib', 'dwc_dbname'),
        query={
            "driver": dwcdriver,
            "Encrypt": "no",
        },
    )

    pguri = URL.create(
        'postgresql',
        username=config.get('dwclib', 'dwcmeta_user'),
        password=config.get('dwclib', 'dwcmeta_pass'),
        host=config.get('dwclib', 'dwcmeta_host'),
        database=config.get('dwclib', 'dwcmeta_dbname'),
    )
else:
    warn(f"Configuration file {configfile} does not exist.", RuntimeWarning)
    dwcuri = None
    pguri = None



from datetime import datetime
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from dwclib.common.db import dwcuri
from dwclib.common.meta import numerics_meta_tz
from dwclib.common.numerics import run_numerics_query
from pandas.api.types import is_list_like


def read_numerics(
    patientids: Union[None, str, List[str]],
    dtbegin: Union[str, datetime],
    dtend: Union[str, datetime],
    labels: Optional[List[str]] = None,
    sublabels: Optional[List[str]] = None,
    pivot: bool = True,
    uri: Optional[str] = None,
) -> pd.DataFrame:
    if not uri:
        uri = dwcuri
    if labels is None:
        labels = []
    if sublabels is None:
        sublabels = []
    if is_list_like(patientids) and len(patientids) == 1:
        patientids = patientids[0]
    df = run_numerics_query(uri, dtbegin, dtend, patientids, labels, sublabels)
    if pivot:
        df = pivot_numerics(df)
    single_patient = patientids is not None and not is_list_like(patientids)
    if single_patient:
        df.columns = df.columns.droplevel(0)
    # Other way to check if there is a single patientid:
    # if len(df.columns.get_level_values(0).drop_duplicates()) == 1:
    return df


def pivot_numerics(df: pd.DataFrame) -> Optional[pd.DataFrame]:
    df = df.dropna(axis=0, how='any', subset=['Value'])
    if not len(df.index):
        numerics_meta_tz
    df = df.pivot_table(
        index=df.index,
        columns=['PatientId', 'SubLabel'],
        values='Value',
        aggfunc=np.nanmax,
    )
    return df



import numpy as np
import pandas as pd


def unfold_row(row: pd.Series, naive_datetime=True) -> pd.Series:
    # XXX unify following line once we use the same unfold_row function for pandas and dask
    begintime = row.name[0] if isinstance(row.name, tuple) else row.name
    basetime = 1000 * begintime.timestamp()
    msperiod = row['SamplePeriod']
    bytesamples = row['WaveSamples']
    calibs = [row[x] for x in ['CAU', 'CAL', 'CSU', 'CSL']]
    noscale = any([x is None for x in calibs])
    if noscale:
        realvals = wave_unfold(bytesamples, 0, 0, 0, 0)
    else:
        realvals = wave_unfold(bytesamples, *calibs)
    # Generate millisecond index
    timestamps = basetime + msperiod * np.arange(len(realvals))
    # Convert to datetime[64]
    timestamps = pd.to_datetime(timestamps, unit='ms', utc=True)
    if naive_datetime:
        timestamps = timestamps.to_numpy(dtype='datetime64[ns]')
    return pd.Series(realvals, index=timestamps)


def wave_unfold(indata: bytes, cau: float, cal: float, csu: int, csl: int):
    if len(indata) % 2:
        indata = indata[:-1]
    int16le = np.dtype('<i2')
    npindata = np.frombuffer(indata, dtype=int16le)
    if any(np.isnan([cau, cal, csu, csl])) or (csu - csl) == 0:
        m = 1
        b = 0
    else:
        m = (cau - cal) / (csu - csl)
        b = cau - m * csu
    outdata = m * npindata.astype(np.float32) + b
    return outdata



from datetime import timedelta
from typing import List, Optional

from dwclib.common.meta import numerics_meta
from dwclib.common.numerics import run_numerics_query
from dwclib.dask.generic import read_data

one_hour = timedelta(hours=1)


def read_numerics(
    patientids,
    dtbegin,
    dtend,
    labels: Optional[List[str]] = None,
    sublabels: Optional[List[str]] = None,
    interval: timedelta = one_hour,
    uri=None,
):
    if labels is None:
        labels = []
    if sublabels is None:
        sublabels = []
    return read_data(
        runner=run_numerics_query,
        meta=numerics_meta,
        dtbegin=dtbegin,
        dtend=dtend,
        uri=uri,
        interval=interval,
        patientids=patientids,
        labels=labels,
        sublabels=sublabels,
        naive_datetime=True,
    )



from typing import List, Optional

import pandas as pd
from dwclib.common.db import pguri
from sqlalchemy import MetaData, Table, asc, create_engine, func, or_, select


def read_patient(*args, **kwargs) -> Optional[pd.Series]:
    res = read_patients(*args, **kwargs, limit=1)
    if len(res):
        return res.iloc[0]


def read_patients(
    patientid: str = None,
    name: str = None,
    ipp: str = None,
    dtbegin: str = None,
    dtend: str = None,
    clinicalunit: str = None,
    bedlabel: str = None,
    wavelabels: Optional[List[str]] = None,
    numericlabels: Optional[List[str]] = None,
    numericsublabels: Optional[List[str]] = None,
    uri: Optional[str] = None,
    limit: Optional[int] = None,
) -> pd.DataFrame:
    """Reads patients from DWCmeta database.

    Retrieves a pandas dataframe consisting of individual patients which match the search criteria.

    Args:
        patientid: A DWC patient identifier
        name: A patient name
        ipp: A patient lifetime identifier
        dtbegin: The beginning of available data samples
        dtend: The end of available data samples
        clinicalunit: The clinical unit the patient belongs to
        bedlabel: The bed / room in which the patient stays
        wavelabels: A list of waveform labels
        numericlabels: A list of numerics labels
        uri: Optional sqlalchemy URI for the database if not provided in the config file
        limit: Maximum number of returned results

    Returns:
        A pandas dataframe with each row corresponding to an individual patient stay, indexed by unique DWC patient identifiers.
    """
    if not uri:
        uri = pguri
    if wavelabels is None:
        wavelabels = []
    if numericlabels is None:
        numericlabels = []
    if numericsublabels is None:
        numericsublabels = []
    q = build_query(
        uri,
        patientid,
        name,
        ipp,
        dtbegin,
        dtend,
        clinicalunit,
        bedlabel,
        wavelabels,
        numericlabels,
        numericsublabels,
    )
    if limit:
        q = q.limit(limit)
    engine = create_engine(uri)
    with engine.connect() as conn:
        df = pd.read_sql(q, conn, index_col='patientid')
    df['data_begin'] = df['data_begin'].to_numpy(dtype='datetime64[ns]')
    df['data_end'] = df['data_end'].to_numpy(dtype='datetime64[ns]')
    return df


def build_query(
    uri,
    patientid,
    name,
    ipp,
    dtbegin,
    dtend,
    clinicalunit,
    bedlabel,
    wavelabels,
    numericlabels,
    numericsublabels,
):
    pgdb = create_engine(uri)
    pgmeta = MetaData(bind=pgdb)

    t_patients = Table('patients', pgmeta, autoload_with=pgdb)
    p = t_patients.c
    t_patientlabels = Table('patientlabels', pgmeta, autoload_with=pgdb)
    pl = t_patientlabels.c

    if name:
        # Build a subquery to allow approximate name search
        subq = select(
            [
                t_patients,
                func.levenshtein(func.lower(p.lastname), func.lower(name)).label(
                    'distance'
                ),
            ]
        ).cte('subq')
        t_patients = subq
        p = t_patients.c

    q = select([t_patients, pl.numericlabels, pl.numericsublabels, pl.wavelabels])
    q = q.select_from(
        t_patients.join(t_patientlabels, pl.patientid == p.patientid, isouter=True)
    )
    if patientid:
        q = q.where(p.patientid == patientid)
    if name:
        q = q.where(p.distance < 3)
        q = q.order_by(asc(p.distance))
    if ipp:
        q = q.where(p.lifetimeid == ipp)
    if bedlabel:
        q = q.where(p.bedlabel == bedlabel)
    if clinicalunit:
        q = q.where(p.clinicalunit == clinicalunit)
    if dtbegin:
        q = q.where(or_(p.data_begin >= dtbegin, p.data_end >= dtbegin))
    if dtend:
        q = q.where(or_(p.data_begin <= dtend, p.data_end <= dtend))
    if numericlabels:
        q = q.where(pl.numericlabels.contains(numericlabels))
    if numericsublabels:
        q = q.where(pl.numericsublabels.contains(numericsublabels))
    if wavelabels:
        q = q.where(pl.wavelabels.contains(wavelabels))
    return q



import pandas as pd
from dwclib.common.meta import waves_meta
from sqlalchemy import MetaData, Table, create_engine, join, select


def run_waves_query(uri, dtbegin, dtend, patientid, labels, naive_datetime=False):
    engine = create_engine(uri)
    q = build_waves_query(engine, dtbegin, dtend, patientid, labels)

    with engine.connect() as conn:
        df = pd.read_sql(q, conn, index_col='TimeStamp')
    engine.dispose()
    if len(df) == 0:
        return waves_meta
    else:
        dtidx = pd.to_datetime(df.index, utc=True)
        df.index = dtidx.to_numpy(dtype='datetime64[ns]') if naive_datetime else dtidx
        return df.astype(waves_meta.dtypes.to_dict(), copy=False)


def build_waves_query(engine, dtbegin, dtend, patientid, labels):
    dbmeta = MetaData(bind=engine)
    wwt = Table('Wave_', dbmeta, schema='_Export', autoload=True, autoload_with=engine)
    wst = Table(
        'WaveSample_', dbmeta, schema='_Export', autoload=True, autoload_with=engine
    )

    ww = select(
        wwt.c.TimeStamp,
        wwt.c.Id,
        wwt.c.Label,
        wwt.c.SamplePeriod,
        wwt.c.CalibrationAbsUpper.label('CAU'),
        wwt.c.CalibrationAbsLower.label('CAL'),
        wwt.c.CalibrationScaledUpper.label('CSU'),
        wwt.c.CalibrationScaledLower.label('CSL'),
    )
    ww = ww.with_hint(wwt, 'WITH (NOLOCK)')
    ww = ww.where(wwt.c.TimeStamp >= dtbegin)
    ww = ww.where(wwt.c.TimeStamp < dtend)
    if labels:
        ww = ww.where(wwt.c.Label.in_(labels))
    ww = ww.cte('Wave')

    ws = select(wst.c.TimeStamp, wst.c.WaveId, wst.c.WaveSamples, wst.c.PatientId)
    ws = ws.with_hint(wst, 'WITH (NOLOCK)')
    ws = ws.where(wst.c.TimeStamp >= dtbegin)
    ws = ws.where(wst.c.TimeStamp < dtend)
    ws = ws.where(wst.c.WaveSamples.is_not(None))
    if patientid:
        ws = ws.where(wst.c.PatientId == patientid)
    ws = ws.cte('WaveSample')

    j = join(ws, ww, ws.c.WaveId == ww.c.Id)
    q = select(
        ws.c.TimeStamp,
        ws.c.PatientId,
        ww.c.Label,
        ws.c.WaveSamples,
        ww.c.SamplePeriod,
        ww.c.CAU,
        ww.c.CAL,
        ww.c.CSU,
        ww.c.CSL,
    ).select_from(j)
    return q



from datetime import datetime
from typing import List, Union

import pandas as pd
from dwclib.common.meta import numerics_meta, numerics_meta_tz
from pandas.api.types import is_list_like
from sqlalchemy import MetaData, Table, create_engine, join, select


def run_numerics_query(
    uri: str,
    dtbegin: Union[str, datetime],
    dtend: Union[str, datetime],
    patientids: Union[None, str, List[str]],
    labels: List[str],
    sublabels: List[str],
    naive_datetime: bool = False,
) -> pd.DataFrame:
    engine = create_engine(uri)
    q = build_numerics_query(engine, dtbegin, dtend, patientids, labels, sublabels)

    with engine.connect() as conn:
        df = pd.read_sql(q, conn, index_col='TimeStamp')
    engine.dispose()
    if len(df) == 0:
        return numerics_meta if naive_datetime else numerics_meta_tz
    else:
        dtidx = pd.to_datetime(df.index, utc=True)
        df.index = dtidx.to_numpy(dtype='datetime64[ns]') if naive_datetime else dtidx
        return df.astype(numerics_meta.dtypes.to_dict(), copy=False)


def build_numerics_query(engine, dtbegin, dtend, patientids, labels, sublabels):
    dbmeta = MetaData(bind=engine)
    nnt = Table(
        'Numeric_', dbmeta, schema='_Export', autoload=True, autoload_with=engine
    )
    nvt = Table(
        'NumericValue_', dbmeta, schema='_Export', autoload=True, autoload_with=engine
    )

    nn = select(nnt.c.TimeStamp, nnt.c.Id, nnt.c.Label, nnt.c.SubLabel)
    nn = nn.with_hint(nnt, 'WITH (NOLOCK)')
    nn = nn.where(nnt.c.TimeStamp >= dtbegin)
    nn = nn.where(nnt.c.TimeStamp < dtend)
    if labels:
        nn = nn.where(nnt.c.Label.in_(labels))
    if sublabels:
        nn = nn.where(nnt.c.SubLabel.in_(sublabels))
    nn = nn.cte('Numeric')

    nv = select(nvt.c.TimeStamp, nvt.c.NumericId, nvt.c.Value, nvt.c.PatientId)
    nv = nv.with_hint(nvt, 'WITH (NOLOCK)')
    nv = nv.where(nvt.c.TimeStamp >= dtbegin)
    nv = nv.where(nvt.c.TimeStamp < dtend)
    nv = nv.where(nvt.c.Value.is_not(None))
    if patientids:
        if is_list_like(patientids):
            nv = nv.where(nvt.c.PatientId.in_(patientids))
        else:
            nv = nv.where(nvt.c.PatientId == patientids)
    nv = nv.cte('NumericValue')

    j = join(nv, nn, nv.c.NumericId == nn.c.Id)
    q = select(
        nv.c.TimeStamp, nv.c.PatientId, nn.c.Label, nn.c.SubLabel, nv.c.Value
    ).select_from(j)
    return q



from datetime import datetime
from functools import lru_cache
from importlib import resources
from typing import List, Optional, Union

import pandas as pd
from pandas.api.types import is_list_like
from sqlalchemy import MetaData, Table, create_engine, func, select

from dwclib import assets
from dwclib.common.db import dwcuri


def read_alerts(
    patientids: Union[None, str, List[str]],
    dtbegin: Union[str, datetime],
    dtend: Union[str, datetime],
    uri: Optional[str] = None,
) -> pd.DataFrame:
    if not uri:
        uri = dwcuri
    if is_list_like(patientids) and len(patientids) == 1:
        patientids = patientids[0]
    df = run_alerts_query(uri, dtbegin, dtend, patientids)
    return df


def run_alerts_query(
    uri: str,
    dtbegin: Union[str, datetime],
    dtend: Union[str, datetime],
    patientids: Union[None, str, List[str]],
) -> pd.DataFrame:
    engine = create_engine(uri)
    q = build_alerts_query(engine, dtbegin, dtend, patientids)

    with engine.connect() as conn:
        df = pd.read_sql(q, conn, index_col='AlertId')
    engine.dispose()

    metadf = load_alerts_meta()
    metadf = metadf[['source_label', 'mdc_alert', 'alert_kind', 'severity']]
    dfr = df.join(metadf, on=['Source', 'Code'])
    return dfr.drop(columns=['Source', 'Code'])


def build_alerts_query(engine, dtbegin, dtend, patientids):
    dbmeta = MetaData(bind=engine)
    at = Table('Alert_', dbmeta, schema='_Export', autoload=True, autoload_with=engine)

    aq = select(
        at.c.AlertId,
        func.min(at.c.TimeStamp).label('begin'),
        func.max(at.c.TimeStamp).label('end'),
        func.max(at.c.Source).label('Source'),
        func.max(at.c.Code).label('Code'),
        func.max(at.c.Label).label('alert_label'),
    )
    aq = aq.with_hint(at, 'WITH (NOLOCK)')
    aq = aq.where(at.c.TimeStamp >= dtbegin)
    aq = aq.where(at.c.TimeStamp < dtend)
    aq = aq.group_by(at.c.AlertId)

    if patientids:
        if is_list_like(patientids):
            aq = aq.where(at.c.PatientId.in_(patientids))
        else:
            aq = aq.where(at.c.PatientId == patientids)
    return aq


@lru_cache
def load_alerts_meta():
    cols = [
        #'alert_text',
        #'mdc_source',
        'physioid',
        'alert_code',
        'mdc_alert',
        'alert_kind',
        'severity',
        'source_label',
        #'source_description',
    ]
    dtypes = {c: 'string' for c in cols}
    dtypes['physioid'] = 'Int64'
    dtypes['alert_code'] = 'Int64'

    with resources.open_text(assets, 'alert_ref.csv') as fd:
        df = pd.read_csv(fd, usecols=cols, index_col=['physioid', 'alert_code'], dtype=dtypes)
    return df



import pandas as pd
from dask.dataframe.utils import make_meta


def build_numerics_meta(naive_datetime=True):
    tz = None if naive_datetime else 'UTC'
    idx = pd.DatetimeIndex([], name='TimeStamp', tz=tz)
    dtypes = {
        'PatientId': 'string',
        'Label': 'string',
        'SubLabel': 'string',
        'Value': 'float32',
    }
    mdf = pd.DataFrame({k: [] for k in dtypes.keys()}, index=idx)
    mdf = mdf.astype(dtype=dtypes)
    return make_meta(mdf)


def build_waves_meta(naive_datetime=True):
    tz = None if naive_datetime else 'UTC'
    idx = pd.DatetimeIndex([], name='TimeStamp', tz=tz)
    dtypes = {
        'PatientId': 'string',
        'Label': 'string',
        'WaveSamples': 'object',
        'SamplePeriod': 'int32',
        'CAU': 'float64',
        'CAL': 'float64',
        'CSU': 'int32',
        'CSL': 'int32',
    }
    mdf = pd.DataFrame({k: [] for k in dtypes.keys()}, index=idx)
    mdf = mdf.astype(dtype=dtypes)
    return make_meta(mdf)


numerics_meta = build_numerics_meta()
numerics_meta_tz = build_numerics_meta(naive_datetime=False)
waves_meta = build_waves_meta()
waves_meta_tz = build_waves_meta(naive_datetime=False)
